
var POOL_NAMES = 
	['i', 'ii', 'iii', 'iv', 'v', 'vi', 'vii', 'viii', 'ix', 'x', 'xi', 'xii' ];

BC_1_8 =   ['AD001', 'AD002', 'AD003', 'AD004', 'AD005', 'AD006', 'AD007', 'AD008'];
BC_9_16 =  ['AD009', 'AD010', 'AD011', 'AD012', 'AD013', 'AD014', 'AD015', 'AD016'];
BC_18_27 = ['AD018', 'AD019', 'AD020', 'AD021', 'AD022', 'AD023', 'AD025', 'AD027'];

// Index Sets = IS
var IS_G1 = {color: 'bg-green',  barcodes: BC_1_8};
var IS_G2 = {color: 'bg-green',  barcodes: BC_9_16};
var IS_G3 = {color: 'bg-green',  barcodes: BC_18_27};

var IS_B1 = {color: 'bg-blue',   barcodes: BC_1_8};
var IS_B2 = {color: 'bg-blue',   barcodes: BC_9_16};
var IS_B3 = {color: 'bg-blue',   barcodes: BC_18_27};

var IS_Y1 = {color: 'bg-yellow', barcodes: BC_1_8};
var IS_Y2 = {color: 'bg-yellow', barcodes: BC_9_16};
var IS_Y3 = {color: 'bg-yellow', barcodes: BC_18_27};


var POOL_SCHEMA =
[
	{
		id: '4by24',
		name: '4 pools × 24 samples',
 		numPools: 4,
 		numWellsPerPool: 24,
 		columnClasses: ['pool-left', 'pool-middle', 'pool-right'],
		barcodeVariants:
			[
			 	{name: 'FULL', indexSets: [IS_G1, IS_B2, IS_Y3, IS_G1, IS_B2, IS_Y3, IS_G1, IS_B2, IS_Y3, IS_G1, IS_B2, IS_Y3]}
			],

		getPoolNumForColumn: function(colNum)
		{
			var poolNum = Math.floor(colNum / 3);
			return poolNum;
		},

 		getClassNameForWell: function(well)
		{
			var c = well.column;
			var cls = 'primary ';
			cls += this.columnClasses[c % 3];
			return cls;
		},
		
		getWellsInPool: function(plate, poolNum)
		{
			var result = [];
			var index = poolNum * this.numWellsPerPool;
			while (result.length < this.numWellsPerPool && index < plate.wells.length)
			{
				result[result.length] = plate.wells[index];
				index++;
			}
			return result;
		}
	
	},
	{
		id: '4by16',
		name: '4 pools × 16 samples',
 		numPools: 4,
 		numWellsPerPool: 16,
 		columnClasses: ['pool-left', 'pool-right'],
		barcodeVariants:
			[
			 	{name: 'A-B-C-A', indexSets: [null, null, IS_G1, IS_G2, IS_B3, IS_B1, IS_Y2, IS_Y3, IS_G1, IS_G2, null, null]},
			 	{name: 'B-C-A-B', indexSets: [null, null, IS_B3, IS_B1, IS_Y2, IS_Y3, IS_G1, IS_G2, IS_B3, IS_B1, null, null]},
			 	{name: 'C-A-B-C', indexSets: [null, null, IS_Y2, IS_Y3, IS_G1, IS_G2, IS_B3, IS_B1, IS_Y2, IS_Y3, null, null]}
			],

		getPoolNumForColumn: function(colNum)
		{
			var poolNum = -1;
			if (colNum >= 2 && colNum <= 9)
			{
				poolNum = Math.floor((colNum - 2) / 2);
			}
			return poolNum;
		},
 		
 		getClassNameForWell: function(well)
		{
			var c = well.column;
			var cls = (c >= 2 && c <= 9) ? 'primary ' : 'secondary ';
			cls += this.columnClasses[c % 2];
			return cls;
		},
		
		getWellsInPool: function(plate, poolNum)
 		{
 			var result = [];
 			var index = poolNum*this.numWellsPerPool + 2 * plate.rows;
			while (result.length < this.numWellsPerPool && index < plate.wells.length)
			{
				result[result.length] = plate.wells[index];
				index++;
			}
 			return result;
 		}
 	},
	{
		id: '5by16',
		name: '5 pools × 16 samples',
		numPools: 5,
		numWellsPerPool: 16,
		columnClasses: ['pool-right', 'pool-left'],
		barcodeVariants:
			[
			 	{name: 'A-B-C-A-B', indexSets: [null, IS_G1, IS_G2, IS_B3, IS_B1, IS_Y2, IS_Y3, IS_G1, IS_G2, IS_B3, IS_B1, null]},
			 	{name: 'B-C-A-B-C', indexSets: [null, IS_B3, IS_B1, IS_Y2, IS_Y3, IS_G1, IS_G2, IS_B3, IS_B1, IS_Y2, IS_Y3, null]},
			 	{name: 'C-A-B-C-A', indexSets: [null, IS_Y2, IS_Y3, IS_G1, IS_G2, IS_B3, IS_B1, IS_Y2, IS_Y3, IS_G1, IS_G2, null]}
			],
	
		getPoolNumForColumn: function(colNum)
		{
			var poolNum = -1;
			if (colNum >= 1 && colNum <= 10)
			{
				poolNum = Math.floor((colNum-1) / 2);
			}
			return poolNum;
		},
 		
		getClassNameForWell: function(well)
		{
			var c = well.column;
			var cls = (c >= 1 && c <= 10) ? 'primary ' : 'secondary ';
			cls += this.columnClasses[c % 2];
			return cls;
		},
		
		getWellsInPool: function(plate, poolNum)
 		{
 			var result = [];
 			var index = poolNum*this.numWellsPerPool+plate.rows;
			while (result.length < this.numWellsPerPool && index < plate.wells.length)
			{
				result[result.length] = plate.wells[index];
				index++;
			}
 			return result;
 		}
	}
];

/**
	Methods for handling pooling layout schemas.
*/
var PoolSchema = function()
{
	var ps = {};
	
	ps.initList = function(list, selectedId)
	{
		if (!selectedId) selectedId = POOL_SCHEMA[0].id;
		var selectedIndex = 0;
		for (var i = 0; i < POOL_SCHEMA.length; i++)
		{
			var schema = POOL_SCHEMA[i];
			var option = new Option(schema.name, schema.id);
			list[list.length] = option;
			if (selectedId == schema.id) selectedIndex = i;
		}
		list.selectedIndex = selectedIndex;
		return POOL_SCHEMA[selectedIndex];
	}
	
	ps.initVariantList = function(list, schema, selectedId)
	{
		if (!selectedId) selectedId = schema.barcodeVariants[0].name;
		var selectedIndex = 0;
		list.length = 0;
		for (var i = 0; i < schema.barcodeVariants.length; i++)
		{
			var variant = schema.barcodeVariants[i];
			var option = new Option(variant.name, variant.name);
			list[list.length] = option;
			if (selectedId == variant.name) selectedIndex = i;
		}
		list.selectedIndex = selectedIndex;
		return schema.barcodeVariants[selectedIndex];
	}
	
	ps.getById = function(schemaId)
	{
		for (var i = 0; i < POOL_SCHEMA.length; i++)
		{
			if (POOL_SCHEMA[i].id == schemaId)
			{
				return POOL_SCHEMA[i];
			}
		}
		return null;
	}
	
	ps.getBarcodeVariantByName = function(schema, barcodeVariant)
	{
		for (var i = 0; i < schema.barcodeVariants.length; i++)
		{
			if (barcodeVariant == schema.barcodeVariants[i].name) 
			{
				return schema.barcodeVariants[i];
			}
		}
		return schema.barcodeVariants[0];
	}
	
	ps.buildPoolTableRow = function(schema, columns, interactive)
	{
		var html = '<th></th>';
		var colNum = 0;
		var poolClass = interactive ? 'pool link' : 'pool';
		var afterPool = false;
		while (colNum < columns)
		{
			var poolNum = schema ? schema.getPoolNumForColumn(colNum) : -1;
			var colspan = 1;
			if (poolNum == -1)
			{
				var className = afterPool ? 'nopool-after-pool' : 'nopool';
				html += '<th class="'+className+'">&nbsp;</th>';
				var afterPool = false;
			}
			else
			{
				while (colNum+colspan < columns)
				{
					if (schema.getPoolNumForColumn(colNum+colspan) != poolNum) break;
					colspan++;
				}
				html += '<th id="pool.' + poolNum + '" colspan="'+colspan+'"';
				html += ' class="'+poolClass+'"';
				html += ' data-pool-num="' + poolNum + '"';
				html += '>'+POOL_NAMES[poolNum]+'</th>';
				afterPool = true;
			}
			colNum += colspan;
		}
		
		var row = document.getElementById('pool-row');
		row.innerHTML = html;
	}
	
	return ps;
}();


var PoolMix = function()
{
	
	var LIMIT_FOR_EXTRA_LARGE_MIX = 1.0;
	var AVAILABLE_VOLUME = 10.0;
	var AVAILABLE_VOLUME_AFTER_SPEEDVAC = 5.0;
	
	var pm = {};
	
	/**
		Round the value to the nearest value with 
		the given number of decimals (default=1).
	*/
	pm.round = function(value, numDecimals)
	{
		if (!numDecimals) numDecimals = 1;
		var pow = Math.pow(10, numDecimals);
		return Math.round(value * pow) / pow;
	}
	
	/**
		Round the value down to the nearest value with 
		the given number of decimals (default=1).
	*/
	pm.floor = function(value, numDecimals)
	{
		if (!numDecimals) numDecimals = 1;
		var pow = Math.pow(10, numDecimals);
		return Math.floor(value * pow) / pow;
	}
	
	/**
		Round the value up to the nearest value with 
		the given number of decimals (default=1).
	*/
	pm.ceil = function(value, numDecimals)
	{
		if (!numDecimals) numDecimals = 1;
		var pow = Math.pow(10, numDecimals);
		return Math.ceil(value * pow) / pow;
	}
	
	/**
		Calculate the volume of a lib to take for a pool given the
		target molarity for the pool and average volume for each lib
		in the pool. The volume is affect by two limits:
		
		1. It may not be lower than LIMIT_FOR_EXTRA_LARGE_MIX. In this case
		   a larger mix is made and only half is used.
		   
		2. It may not be higher than AVAILABLE_VOLUME or AVAILABLE_VOLUME_AFTER_SPEEDVAC.
		
		The volume is rounded to one decimal in µl.
	*/
	pm.calculateLibVolume = function(lib, targetMolarity, targetVolume, mixingStrategy)
	{
		var maxVolume = lib.speedVacConc != null ? AVAILABLE_VOLUME_AFTER_SPEEDVAC : AVAILABLE_VOLUME;
		if (mixingStrategy == 'fixed' && maxVolume > targetVolume)
		{
			maxVolume = targetVolume;
		}
		var volume = pm.round(targetVolume * targetMolarity / lib.molarity);
		var mixFactor = 1;
		if (volume < LIMIT_FOR_EXTRA_LARGE_MIX)
		{
			mixFactor = 2;
			// Re-calculate the volume to get the rounding correct (eg. to avoid dubling rounding errors)
			// round UP this time to avoid storing something that is less than 2nM
			volume = pm.ceil(mixFactor * targetVolume * targetMolarity / lib.molarity) / mixFactor;
		}
		
		lib.maxVolume = maxVolume;
		lib.volume = volume;
		lib.mixFactor = mixFactor;
		lib.eb = null;
		return volume;
	}
	
	/**
		Calculate the amount of EB to mix with each of the libs. If mixing
		strategy is 'fixed' all libraries are mixed to the target volume,
		otherwise a dynamic strategy that prioritize the final pool molarity
		is used:
		
		1. The total amount of EB needed to achive the target molarity
		   in the pool is calculated. 
		   
		2. Libs with a mixFactor > 1 are always mixed to the given target 
		   molarity since the saved lib should be standardized
		   
		3. The remaining EB volume is equally split among the rest of the
		   libs. To avoid rounding errors for the pool as a total the
		   volume is rounded down and then any remaining EB is added to
		   the last lib.
	*/
	pm.calculateEbForLibs = function(libs, targetMolarity, targetVolume, mixingStrategy)
	{
		if (mixingStrategy == 'fixed')
		{
			// Fixed strategy
			// All libs are mixed with EB to target volume
			for (var libNo = 0; libNo < libs.length; libNo++)
			{
				var lib = libs[libNo];
				lib.eb = targetVolume - lib.volume;
			}
		}
		else
		{
			// Dynamic strategy
			// First, calculate total volume of all libs 
			// and their contribution to the total amount of DNA in the pool
			var totalLibVolume = 0;
			var totalLibAmount = 0;
			var fixedEbVolume = 0;
			var fixedLibs = 0;
			for (var libNo = 0; libNo < libs.length; libNo++)
			{
				var lib = libs[libNo];
				var vol = lib.maxVolume < lib.volume ? lib.maxVolume : lib.volume;
				totalLibVolume += vol;
				totalLibAmount += vol * lib.molarity;
				
				// Always mix to target molarity when doing extra large mix
				if (lib.mixFactor > 1)
				{
					var eb = targetVolume - lib.volume;
					fixedEbVolume += eb;
					lib.eb = eb;
					fixedLibs++;
				}
			}

			
			// Calculate the total volume needed to get to the
			// target molarity for the pooled libs
			var totalPoolVolume = totalLibAmount / targetMolarity;
			var totalEbVolume = totalPoolVolume - totalLibVolume - fixedEbVolume;

			// Divide EB volume among the libs (rounded to 1 decimal)
			var ebPerLib = pm.floor(totalEbVolume / (libs.length - fixedLibs), 1);
			var usedEb = 0;
			var adjustableLib;
			for (var libNo = 0; libNo < libs.length; libNo++)
			{
				var lib = libs[libNo];
				if (!lib.mixFactor || lib.mixFactor == 1)
				{
					lib.eb = ebPerLib;
					usedEb += ebPerLib;
					adjustableLib = lib;
				}
			}
			
			// If the used EB doesn't add up to the total amount needed (due to rounding)
			// add the extra amount to the last lib
			if (Math.abs(totalEbVolume - usedEb) > 0.1 && adjustableLib)
			{
				adjustableLib.eb += totalEbVolume - usedEb;
			}
		}
		
	}
	
	pm.calculateFinalPoolInfo = function(libs, mixingStrategy)
	{
		var info = {};
		
		var libVolume = 0;
		var libAmount = 0;
		var ebVolume = 0;
		var ebFinalMix = 0;
		for (var libNo = 0; libNo < libs.length; libNo++)
		{
			var lib = libs[libNo];
			var vol = lib.maxVolume < lib.volume ? lib.maxVolume : lib.volume;
			var eb = Math.max(0, lib.eb);
			libVolume += vol;
			ebVolume += eb;
			if (mixingStrategy == 'dynamic' && (!lib.mixFactor || lib.mixFactor == 1)) 
			{
				ebFinalMix += eb; 
			}
			libAmount += vol * lib.molarity;
		}
		
		info.volume = libVolume;
		info.amount = libAmount;
		info.eb = ebVolume;
		info.ebFinalMix = ebFinalMix;
		if (ebVolume < 0)
		{
			info.totalVolume = libVolume;
		}
		else
		{
			info.totalVolume = libVolume + ebVolume;
		}
		info.molarity = libAmount / info.totalVolume;
		
		return info;
	}
	
	return pm;
}();
